""" Tests for building image ingestion manifest """

from pathlib import Path
from typing import List

# pylint: disable=C0103, E0401, E0402, R0914, R1721, W0621

import json
import shutil
import tarfile

from ingest_envoy.ingestion_manifest import IngestionManifestBuilder
from ingest_envoy.schema import AbstractTextFile
from ingest_envoy.utilities import Telescope, AncillaryProductType, ScienceProductType

# ingest_path is NOT unused. IJ is dumb.
from .conftest import (
    ingest_path,
    populate_fake_image_ingest_path,
    IMG_MANIFEST_INPUT_FILENAMES,
    CASA_BYPRODUCTS,
)

from ingest_envoy.manifest_components import (
    WEBLOG_FILENAME,
    ManifestParameters,
    ParamsKey,
    OutputGroup,
    OutputScienceProduct,
    AncillaryProduct,
    InputGroup,
    InputScienceProduct,
    MANIFEST_FILENAME,
    TARFILE_EXT,
)


def test_parameters_json_well_formed(ingest_path: Path):
    """
    Make sure we get expected ManifestParameters for an image.

    :return:
    """
    params = manifest_parameters(ingest_path)

    params_json = params.to_json()
    # Only difference from EVLA CAL manifest is additional_metadata
    assert params_json[ParamsKey.ADDITIONAL_METADATA.value] == str(params.additional_metadata)

    # gotta be able to dump it; test will fail if dump runs into trouble
    json.dumps(params_json)

    shutil.rmtree(ingest_path)


def test_output_science_prods_built_properly(ingest_path: Path):
    """
    Output science products for an image ingestion contain ancillary products.
    Let's make sure we can build one of these OSPs.
    :return:
    """
    aps = rms_ancillary_prods(ingest_path)
    osp = osp_ql_with_ancillaries(ingest_path, aps)

    # check the output science product
    assert osp.type == AncillaryProductType.QUICKLOOK_IMAGE
    assert osp.filename.endswith(
        "VLASS2.1.ql.T08t09.J055438-113000.10.2048.v1.I.iter1.image.pbcor.tt0.subim.fits"
    )

    # check the ancillary products belonging to the science product
    assert len(osp.ancillary_products) == 2
    osp_ap = osp.ancillary_products[0]
    assert osp_ap == aps[0]
    aps_json = [ap.to_json() for ap in aps]
    # if we can't dump it, we know it's no good
    ap_str = json.dumps(aps_json)

    # confirm the output science products "look" right
    osp_json = osp.to_json()
    assert osp_json["type"] == osp.type.value
    assert osp_json["filename"] == osp.filename

    op_str = json.dumps(osp_json)
    assert ap_str in op_str

    shutil.rmtree(ingest_path)


def test_output_group_made_correctly(ingest_path: Path):
    """
    Does the output group produced by the manifest builder contain everything it should?

    :param ingest_path: staging source dir
    :return:
    """

    # make some fake science-product-associated ancillary products
    sp_aps = rms_ancillary_prods(ingest_path)
    # make an output science product with 'em
    osp = osp_ql_with_ancillaries(ingest_path, sp_aps)

    # make the other ancillary products
    aps = separate_aps(ingest_path)

    # make the output group using them....
    output_group = OutputGroup(science_products=[osp], ancillary_products=aps)
    # ... and confirm it contains what we expected.
    assert output_group.science_products[0] == osp
    assert output_group.ancillary_products == aps

    shutil.rmtree(ingest_path)


def test_output_group_json_well_formed(ingest_path: Path):
    """
    Does the output group produced by the manifest builder "look right"?

    :param ingest_path: staging source dir
    :return:
    """

    output_group = build_output_group(ingest_path)
    expected_og_json = output_group.to_json()

    sps = output_group.science_products

    # since these are image science products, we expect there to be
    # ancillary image products
    for sp in sps:
        assert len(sp.ancillary_products) > 0
    expected_sp_json = [sp.to_json() for sp in sps]

    other_aps = separate_aps(ingest_path)
    expected_aps_json = [ap.to_json() for ap in other_aps]

    # if any of the JSON isn't well formed, raise a stink....
    og_dump = json.dumps(expected_og_json)
    aps_dump = json.dumps(expected_aps_json)
    sps_dump = json.dumps(expected_sp_json)
    # ... and make sure each of the dumps contains what it should.
    assert sps_dump in og_dump
    assert aps_dump in og_dump
    assert aps_dump not in sps_dump

    shutil.rmtree(ingest_path)


def test_ingestion_artifacts_tar_correct(ingest_path: Path):
    """
    Do we build the correct ingestion_artifacts tar for for standard imaging ingestion?

    :return:
    """
    # fill the ingestion path with fake files
    populate_fake_image_ingest_path(ingest_path)

    locator = "uid://evla/calibration/3dfa528b-9870-46c9-a200-131dbac701cc"
    addl_md = AbstractTextFile(filename="image_metadata_2021_05_21_T10_17_19.180.json", content="")

    # manifest maker, manifest maker, make me a manifest
    IngestionManifestBuilder(
        staging_source_dir=ingest_path,
        sp_type=ScienceProductType.IMAGE.value,
        locator=locator,
        telescope=Telescope.EVLA.value,
        additional_metadata=addl_md,
    ).build()

    # get ingestion artifacts
    artifacts_file = [file for file in ingest_path.glob("ingestion_artifacts*.tar")][0]
    with tarfile.open(artifacts_file, "r") as tar:
        # confirm that contains has as many files as we expect...
        members = tar.getmembers()
        assert len(members) == len(IMG_MANIFEST_INPUT_FILENAMES) + len(CASA_BYPRODUCTS) + 2

    # ...and that they include the manifest and CASA byproducts
    filenames = [Path(tar_info.name).name for tar_info in members]
    manifest_found = False
    casa_byproducts = []
    for filename in filenames:
        if filename == MANIFEST_FILENAME:
            manifest_found = True
        elif filename in CASA_BYPRODUCTS:
            casa_byproducts.append(filename)

    assert manifest_found
    assert len(casa_byproducts) == len(CASA_BYPRODUCTS)

    shutil.rmtree(ingest_path)


def test_creates_expected_manifest(ingest_path: Path):
    """
    Did the image ingestion manifest builder make the manifest we expected?

    :return:
    """

    # fill the ingestion path with fake files
    populate_fake_image_ingest_path(ingest_path)

    locator = "uid://evla/calibration/3dfa528b-9870-46c9-a200-131dbac701cc"
    addl_md_file = ingest_path / "image_metadata_2021_05_21_T10_17_19.180.json"
    addl_md = AbstractTextFile(filename=str(addl_md_file), content="")

    # build us a manifest in the ingest_path using this locator and additional metadata
    builder = IngestionManifestBuilder(
        staging_source_dir=ingest_path,
        sp_type=ScienceProductType.IMAGE.value,
        locator=locator,
        telescope=Telescope.EVLA.value,
        additional_metadata=addl_md,
    )
    manifest, _ = builder.build()

    expected_params = manifest_parameters(ingest_path)
    actual_params = manifest.parameters
    assert actual_params == expected_params

    assert manifest.input_group == InputGroup(science_products=[InputScienceProduct(locator)])

    # make the output group the manifest should have
    osp = osp_ql_with_ancillaries(ingest_path, rms_ancillary_prods(ingest_path))
    aps = separate_aps(ingest_path)
    expected_output_group = OutputGroup(science_products=[osp], ancillary_products=aps)

    # the output groups won't be -exactly- the same since the ingestion artifacts
    # filenames will differ, but we can compare the parts of them
    # that should match
    actual_output_group = manifest.output_group
    actual_sci_prods = actual_output_group.science_products
    assert actual_sci_prods == expected_output_group.science_products

    expected_aps = expected_output_group.ancillary_products
    actual_aps = actual_output_group.ancillary_products

    assert len(actual_aps) == len(expected_aps)

    # compare the weblogs
    expected_weblog = [
        ap for ap in expected_aps if ap.type == AncillaryProductType.PIPELINE_WEBLOG_TYPE
    ][0]
    actual_weblog = [
        ap for ap in actual_aps if ap.type == AncillaryProductType.PIPELINE_WEBLOG_TYPE
    ][0]
    assert actual_weblog == expected_weblog

    # compare pipeline artifacts
    expected_pip_art = [
        ap for ap in expected_aps if ap.type == AncillaryProductType.PIPELINE_ARTIFACTS
    ][0]
    actual_pip_art = [
        ap for ap in actual_aps if ap.type == AncillaryProductType.PIPELINE_ARTIFACTS
    ][0]
    assert actual_pip_art == expected_pip_art
    expected_tar = Path(expected_pip_art.filename)
    actual_tar = Path(actual_pip_art.filename)
    assert actual_tar.exists()
    assert actual_tar == expected_tar

    # the ingestion artifacts files won't have exactly the same name, because the filename
    # is created on the fly using the current timestamp, but they should be identical otherwise
    expected_ing_art = [
        ap for ap in expected_aps if ap.type == AncillaryProductType.INGESTION_ARTIFACTS
    ][0]
    assert "ingestion_artifacts" in expected_ing_art.filename
    actual_ing_art = [
        ap for ap in actual_aps if ap.type == AncillaryProductType.INGESTION_ARTIFACTS
    ][0]
    assert "ingestion_artifacts" in actual_ing_art.filename
    expected_tar = Path(expected_ing_art.filename)
    actual_tar = Path(actual_ing_art.filename)
    assert actual_tar.exists()
    assert actual_tar.parent == expected_tar.parent

    shutil.rmtree(ingest_path)


def test_writes_expected_output_files(ingest_path: Path):
    """
    Did the image ingestion manifest builder produce the output file(s) we expect?

    :param ingest_path:
    :return:
    """
    populate_fake_image_ingest_path(ingest_path)
    manifest, manifest_file = IngestionManifestBuilder(
        telescope=Telescope.EVLA.value,
        staging_source_dir=ingest_path,
        locator="uid://evla/image/kiss-me-Im_Elvish",
        sp_type=ScienceProductType.IMAGE.value,
    ).build()
    assert manifest_file
    assert manifest_file.name == MANIFEST_FILENAME
    assert manifest

    ingestion_files = [file for file in ingest_path.iterdir()]

    # In addition to the manifest, the science products tar and the ingestion artifact,
    # we expect to find the additional metadata .json, two .fits image files, a thumbnail.png,
    # CASA byproducts, and the random tar we tossed in
    assert len(ingestion_files) == 13
    files_accounted_for = []

    fits = [file for file in ingestion_files if file.name.endswith(".fits")]
    assert len(fits) == 2
    for file in fits:
        files_accounted_for.append(file)

    thumbnails = [file for file in ingestion_files if file.name.endswith(".png")]
    assert len(thumbnails) == 1
    files_accounted_for.append(thumbnails[0])

    casa_byproducts = [file for file in ingestion_files if file.name in CASA_BYPRODUCTS]
    assert len(casa_byproducts) == len(CASA_BYPRODUCTS)
    for file in casa_byproducts:
        files_accounted_for.append(file)

    manifests = [file for file in ingestion_files if file.name == MANIFEST_FILENAME]
    assert len(manifests) == 1
    files_accounted_for.append(manifests[0])

    inputs = [file for file in ingestion_files if file.name in IMG_MANIFEST_INPUT_FILENAMES]
    assert len(inputs) == len(IMG_MANIFEST_INPUT_FILENAMES)
    for file in inputs:
        files_accounted_for.append(file)

    artifacts = [file for file in ingestion_files if is_ingestion_artifact(file)]
    assert len(artifacts) == 1
    files_accounted_for.append(artifacts[0])

    assert len(set(files_accounted_for)) == len(ingestion_files) - 1

    shutil.rmtree(ingest_path)


# -----------------------------
#  U  T  I  L  I  T  I  E  S
# -----------------------------


def manifest_parameters(ingest_path: Path) -> ManifestParameters:
    """
    Build a ManifestParameters for our tests

    :return: the manifest parameters we're expecting
    """
    addl_md_path = ingest_path / "image_metadata_2021_05_21_T10_17_19.180.json"
    addl_md = AbstractTextFile(filename=str(addl_md_path), content="")

    params = ManifestParameters(
        telescope=Telescope.EVLA,
        reingest=False,
        ngas_ingest=True,
        calibrate=False,
        staging_source_dir=ingest_path,
        additional_metadata=AbstractTextFile(filename=addl_md.filename, content=""),
    )

    return params


def rms_ancillary_prods(staging_source_dir: Path) -> List[AncillaryProduct]:
    """
    Build an AncillaryProduct to use in a test.

    :return: the product we're expecting
    """
    file1 = (
        staging_source_dir
        / "VLASS2.1.ql.T08t09.J055438-113000.10.2048.v1.I.iter1.image.pbcor.tt0.rms.subim.fits"
    )
    ap1 = AncillaryProduct(
        type=AncillaryProductType.QUICKLOOK_RMS_IMAGE,
        filename=str(file1),
    )
    file2 = (
        staging_source_dir
        / "VLASS2.1.ql.T08t09.J055438_113000.10.2048.v1.I.iter1.image.pbcor.tt0.subim.png"
    )
    ap2 = AncillaryProduct(
        type=AncillaryProductType.THUMBNAIL_IMG,
        filename=str(file2),
    )

    return [ap1, ap2]


def osp_ql_with_ancillaries(
    staging_source_dir: Path, aps: List[AncillaryProduct]
) -> OutputScienceProduct:
    """
    Build an OutputScienceProduct for a test

    :return: the product we're expecting
    """

    sp_file = (
        staging_source_dir
        / "VLASS2.1.ql.T08t09.J055438-113000.10.2048.v1.I.iter1.image.pbcor.tt0.subim.fits"
    )
    return OutputScienceProduct(
        type=AncillaryProductType(AncillaryProductType.QUICKLOOK_IMAGE),
        filename=str(sp_file),
        ancillary_products=aps,
    )


def separate_aps(staging_source_dir: Path) -> List[AncillaryProduct]:
    """
    Build a list of AncillaryProducts for our tests

    :return: the products we're expecting
    """
    file1 = staging_source_dir / WEBLOG_FILENAME
    ap1 = AncillaryProduct(type=AncillaryProductType.PIPELINE_WEBLOG_TYPE, filename=str(file1))

    file2 = staging_source_dir / "VLASS2.1.ql.T08t09.J055438-113000.10.2048.v1.tar"
    ap2 = AncillaryProduct(
        type=AncillaryProductType.PIPELINE_ARTIFACTS,
        filename=str(file2),
    )
    file3 = staging_source_dir / "ingestion_artifacts_2021_05_21_T10_17_19.275.tar"
    ap3 = AncillaryProduct(
        type=AncillaryProductType.INGESTION_ARTIFACTS,
        filename=str(file3),
    )
    return [ap1, ap2, ap3]


def build_output_group(staging_source_dir: Path) -> OutputGroup:
    """
    Build an OutputGroup for our tests

    :return: the output group we're expecting
    """
    osp = osp_ql_with_ancillaries(staging_source_dir, rms_ancillary_prods(staging_source_dir))

    other_aps = separate_aps(staging_source_dir)
    ap_list = other_aps

    return OutputGroup(science_products=[osp], ancillary_products=ap_list)


def is_ingestion_artifact(file: Path) -> bool:
    """
    We ought to have an ingestion artifact named like "ingestion_artifacts_*.tar"

    :param file: some tar we found in staging source dir
    :return:
    """
    return file.name.startswith("ingestion_artifacts") and file.name.endswith(TARFILE_EXT)
